#ifndef PYTHONIC_INCLUDE_OPERATOR_EQ__HPP
#define PYTHONIC_INCLUDE_OPERATOR_EQ__HPP

#include "pythonic/include/operator_/eq.hpp"

PYTHONIC_NS_BEGIN

namespace operator_
{

  USING_FUNCTOR(__eq__, eq);
}
PYTHONIC_NS_END

#endif
