#ifndef PYTHONIC_INCLUDE_OPERATOR_ISUB__HPP
#define PYTHONIC_INCLUDE_OPERATOR_ISUB__HPP

#include "pythonic/include/operator_/isub.hpp"

PYTHONIC_NS_BEGIN

namespace operator_
{
  USING_FUNCTOR(__isub__, isub);
}
PYTHONIC_NS_END

#endif
