#ifndef PYTHONIC_INCLUDE_OPERATOR_SUB__HPP
#define PYTHONIC_INCLUDE_OPERATOR_SUB__HPP

#include "pythonic/include/operator_/sub.hpp"

PYTHONIC_NS_BEGIN

namespace operator_
{
  USING_FUNCTOR(__sub__, sub);
}
PYTHONIC_NS_END

#endif
