#ifndef PYTHONIC_OPERATOR_IMUL_HPP
#define PYTHONIC_OPERATOR_IMUL_HPP

#include "pythonic/include/operator_/imul.hpp"

#define OPERATOR_NAME imul
#define OPERATOR_SYMBOL *
#define OPERATOR_ISYMBOL *=

#include "pythonic/operator_/icommon.hpp"

#endif
