



set "SRC_DIR="
IF %ERRORLEVEL% NEQ 0 exit /B 1
pythran -v dprod.py
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -c "import dprod"
IF %ERRORLEVEL% NEQ 0 exit /B 1
pythran -v simple_numexpr.py -DUSE_XSIMD -march=native
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -c "import simple_numexpr"
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
